#!/usr/bin/env python
# -*- coding: utf-8 -*-
# @Project      : AI.  @by PyCharm
# @File         : OpenAIEmbeddings
# @Time         : 2023/7/11 18:40
# @Author       : betterme
# @WeChat       : meutils
# @Software     : PyCharm
# @Description  :

import langchain
from langchain.embeddings import OpenAIEmbeddings as _OpenAIEmbeddings

from meutils.pipe import *
from chatllm.llmchain.utils import get_api_key


class OpenAIEmbeddings(_OpenAIEmbeddings):
    """多key多线程"""
    get_api_key: Callable[[int], List[str]] = get_api_key
    pre_fn: Optional[Callable[[str], str]] = None

    # class Config:
    #     """Configuration for this pydantic object."""
    #
    #     allow_population_by_field_name = True
    def embed_documents(
        self,
        texts: List[str],
        chunk_size: Optional[int] = 0,
    ) -> List[List[float]]:
        if self.pre_fn: texts = texts | xmap_(self.pre_fn)

        n = int(np.ceil(len(texts) / self.chunk_size))
        api_key_set = self.get_api_key(n=n)

        max_workers = np.clip(len(api_key_set), 1, 32).astype(int)  # 最大线程数
        if max_workers > 1:
            embeddings_map = {}
            for i, api_key in enumerate(api_key_set):
                kwargs = self.dict().copy()
                kwargs.pop('get_api_key', None)  # not permitted
                kwargs['openai_api_key'] = api_key
                embeddings_map[i] = _OpenAIEmbeddings(**kwargs)  # 可以用 OpenAIEmbeddings

            if langchain.debug:
                logger.info([e.openai_api_key for e in embeddings_map.values()])
                logger.info(f"Maximum concurrency: {max_workers * self.chunk_size}")

            def __embed_documents(arg):
                idx, texts = arg
                embeddings = embeddings_map.get(idx % max_workers, 0)
                return embeddings.embed_documents(texts)

            return (
                texts | xgroup(self.chunk_size)
                | xenumerate
                | xThreadPoolExecutor(__embed_documents, max_workers)
                | xchain_
            )

        return super().embed_documents(texts)


if __name__ == '__main__':
    e = OpenAIEmbeddings(chunk_size=5)

    e.get_api_key = partial(get_api_key, n=2)
    # e.openai_api_key = 'xxx'
    print(e.get_api_key())
    print(e.openai_api_key)
    print(e.embed_documents(['x'] * 6))
    print(e.embed_query('x'))



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# @Project      : AI.  @by PyCharm
# @File         : ElasticsearchStore
# @Time         : 2023/9/18 15:40
# @Author       : betterme
# @WeChat       : meutils
# @Software     : PyCharm
# @Description  :

from langchain.docstore.document import Document
from langchain.vectorstores import ElasticsearchStore as _ElasticsearchStore, VectorStore

from meutils.pipe import *
from meutils.decorators.retry import retrying


class ElasticsearchStore(_ElasticsearchStore):

    def similarity_search(
        self,
        query: str,
        k: int = 4,
        filter: Optional[List[dict]] = None,
        threshold: float = 0.5,
        **kwargs: Any,
    ) -> List[Document]:

        docs_and_scores = self._search(query=query, k=k, filter=filter, **kwargs)
        docs = []
        for doc, score in docs_and_scores:
            if score > threshold:
                doc.metadata['score'] = round(score, 2)
                docs.append(doc)
        return docs

    @retrying
    def _search(
        self,
        query: Optional[str] = None,
        k: int = 4,
        query_vector: Union[List[float], None] = None,
        fetch_k: int = 50,
        fields: Optional[List[str]] = None,
        filter: Optional[List[dict]] = None,
        custom_query: Optional[Callable[[Dict, Union[str, None]], Dict]] = None,
    ) -> List[Tuple[Document, float]]:
        return super()._search(query, k, query_vector, fetch_k, fields, filter, custom_query)

    @staticmethod
    def connect_to_elasticsearch(
        *,
        es_url: Optional[str] = None,
        cloud_id: Optional[str] = None,
        api_key: Optional[str] = None,
        username: Optional[str] = None,
        password: Optional[str] = None,
    ):
        try:
            import elasticsearch
        except ImportError:
            raise ImportError(
                "Could not import elasticsearch python package. "
                "Please install it with `pip install elasticsearch`."
            )

        if es_url and cloud_id:
            raise ValueError(
                "Both es_url and cloud_id are defined. Please provide only one."
            )

        connection_params: Dict[str, Any] = {}

        if es_url:
            connection_params["hosts"] = [es_url]
        elif cloud_id:
            connection_params["cloud_id"] = cloud_id
        else:
            raise ValueError("Please provide either elasticsearch_url or cloud_id.")

        if api_key:
            connection_params["api_key"] = api_key
        elif username and password:
            connection_params["basic_auth"] = (username, password)

        #########################新增#########################
        # 第一次失败，第二次成功，需要加重试逻辑
        # 在做任何操作之前，先进行嗅探
        sniff_on_start = True

        # 节点没有响应时，进行刷新，重新连接
        sniff_on_node_failure = True

        # 每 60 秒刷新一次
        min_delay_between_sniffing = 60
        ######################################################

        es_client = elasticsearch.Elasticsearch(
            **connection_params,
            sniff_on_start=sniff_on_start,
            sniff_on_node_failure=sniff_on_node_failure,
            min_delay_between_sniffing=min_delay_between_sniffing
        )
        try:
            es_client.info()
        except Exception as e:
            logger.error(f"Error connecting to Elasticsearch: {e}")
            raise e

        return es_client



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# @Project      : AI.  @by PyCharm
# @File         : MultiPrompt
# @Time         : 2023/10/25 14:26
# @Author       : betterme
# @WeChat       : meutils
# @Software     : PyCharm
# @Description  : https://www.langchain.com.cn/modules/chains/examples/multi_prompt_router

from meutils.pipe import *

from langchain.chains.router import MultiPromptChain


class PromptBase(BaseModel):
    """
    prompt_infos = [
    {
        "name": "physics",
        "description": "Good for answering questions about physics",
        "prompt_template": You are a very good mathematician. Here is a question: {input}
    }
]

    """
    name: str
    description: str
    prompt_template: str




from langchain.chains import APIChain



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# @Project      : AI.  @by PyCharm
# @File         : 4v
# @Time         : 2023/11/15 17:49
# @Author       : betterme
# @WeChat       : meutils
# @Software     : PyCharm
# @Description  :

from meutils.pipe import *
from pydantic import validator
from meutils.io.image import image_to_base64
from openai import OpenAI

client = OpenAI()
import langchain

class TextContent(BaseModel):
    type: str = "text"
    text: str


class ImageURL(BaseModel):
    url: str
    detail: Literal['low', 'high', 'auto'] = 'auto'

    @validator('url')
    def validate(cls, url):
        if not url.startswith('http') and Path(url).is_file():
            url = image_to_base64(url, for_image_url=True)
        return url


class ImageContent(BaseModel):
    type: str = "image_url"
    image_url: ImageURL


response = client.chat.completions.create(
    model="gpt-4-vision-preview",
    messages=[
        {
            "role": "user",
            "content": [
                {"type": "text", "text": "What’s in this image?"},
                {
                    "type": "image_url",
                    "image_url": {
                        "url": "https://upload.wikimedia.org/wikipedia/commons/thumb/d/dd/Gfp-wisconsin-madison-the-nature-boardwalk.jpg/2560px-Gfp-wisconsin-madison-the-nature-boardwalk.jpg",
                        "detail": "high"
                    },
                },
            ],
        }
    ],
    max_tokens=300,
)

# print(response.choices[0].message.content)
if __name__ == '__main__':
    # a = [TextContent(text="What’s in this image?"), ImageContent(image_url=ImageURL(url='http://xxx'))]
    # print(a)

    image_url = ImageURL(url='xxx')
    print(ImageContent(image_url=image_url))

    print(image_url)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# @Project      : AI.  @by PyCharm
# @File         : protocol
# @Time         : 2023/11/29 16:45
# @Author       : betterme
# @WeChat       : meutils
# @Software     : PyCharm
# @Description  :

from meutils.pipe import *


class Cookie(BaseModel):
    name: str
    value: str
    domain: str
    path: str
    expires: float
    httpOnly: bool
    secure: bool
    sameSite: str


class StorageItem(BaseModel):
    name: str
    value: str


class Origin(BaseModel):
    origin: str
    localStorage: List[StorageItem]


class State(BaseModel):
    """
        state = State(**json.loads(Path('state.json').read_text()))
        refresh_token = state.origins[0].localStorage[1].value
        access_token = state.origins[0].localStorage[2].value
    """
    cookies: List[Cookie]
    origins: List[Origin]


class SearchMessage(BaseModel):
    need2search: bool
    search_target: str
    search_language: str


class EventData(BaseModel):
    """
    data: {"content":"1+1","event":"req","group_id":"cljdett0tc12f8jcpc7g","id":"cljdett0tc12f8jcpc8g","refs":[]}
    data: {"event":"resp","group_id":"cljdett0tc12f8jcpc7g","id":"cljdett0tc12f8jcpc90"}
    data: {"event":"debug","message":{"need2search":false,"search_target":"","search_language":""}}
    data: {"event":"debug","message":{"need2search":true,"search_target":"南京今天天气","search_language":"zh-cn"}}
    data: {"event":"search_plus","msg":{"type":"start"}}
    data: {"event":"search_plus","msg":{"successNum":1,"title":"南京天气","type":"get_res","url":"http://m.nmc.cn/publish/forecast/AJS/nanjing.html"}}

    data: {"event":"cmpl","text":" "}
    data: {"event":"cmpl","text":"1"}
    data: {"event":"cmpl","text":"å "}
    data: {"event":"done"}
    data: {"event":"cmpl","text":"1"}
    data: {"event":"cmpl","text":"ç­"}
    data: {"event":"cmpl","text":"äº"}
    data: {"event":"cmpl","text":"2"}
    data: {"event":"cmpl","text":"ã"}
    data: {"event":"all_done"}
    """
    event: Optional[Literal["req", "resp", "debug", "search_plus", "cmpl", "done", "all_done"]] = None
    content: Optional[str] = None

    id: Optional[str] = None
    group_id: Optional[str] = None

    message: Optional[SearchMessage] = None
    msg: Optional[dict] = None
    refs: List[str] = []

    text: str = ''

    # json_str = {"messages": [{"role": "system", "content": "你现在扮演的角色是初中老师"},
    #                          {"role": "user", "content": "你现在扮演的角色是？"}], "refs": [], "use_search": False}


if __name__ == '__main__':
    state = State(**json.loads(Path('state.json').read_text()))
    refresh_token = state.origins[0].localStorage[1].value
    access_token = state.origins[0].localStorage[2].value



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# @Project      : AI.  @by PyCharm
# @File         : pdf
# @Time         : 2023/6/30 18:58
# @Author       : betterme
# @WeChat       : meutils
# @Software     : PyCharm
# @Description  :

from langchain.document_loaders.parsers.pdf import BaseBlobParser, Blob, Document
from langchain.document_loaders.pdf import PyMuPDFLoader as _PyMuPDFLoader

from meutils.pipe import *
from meutils.fileparser import stream_parser


class PyMuPDFLoader(_PyMuPDFLoader):
    """Loader that uses PyMuPDF to load PDF files."""

    def __init__(self, file_path: Any) -> None:
        """Initialize with a file path."""
        try:
            import fitz  # noqa:F401
        except ImportError:
            raise ImportError(
                "`PyMuPDF` package not found, please install it with "
                "`pip install pymupdf`"
            )

        self.file_path = file_path

    def load(self, **kwargs: Optional[Any]) -> List[Document]:
        if (
            isinstance(self.file_path, (str, os.PathLike))
            and len(self.file_path) < 256
            and Path(self.file_path).is_file()
        ):
            return _PyMuPDFLoader(self.file_path).load()  # 按页

        filename, file_stream = stream_parser(self.file_path)
        text, metadata = self.get_text(file_stream)
        return [Document(page_content=text, metadata=metadata)]

    @staticmethod
    def get_text(stream):
        import fitz
        doc = fitz.Document(stream=stream)
        return '\n'.join(page.get_text().strip() for page in doc), {'total_pages': len(doc), **doc.metadata}


if __name__ == '__main__':
    f = open('/Users/betterme/PycharmProjects/AI/ChatLLM/data/中职职教高考政策解读.pdf')
    print(PyMuPDFLoader(f).load())



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# @Project      : AI.  @by PyCharm
# @File         : quote
# @Time         : 2023/11/1 10:46
# @Author       : betterme
# @WeChat       : meutils
# @Software     : PyCharm
# @Description  :

from langchain.prompts import SystemMessagePromptTemplate, HumanMessagePromptTemplate, ChatPromptTemplate

system = '''
你的背景知识:
"""
{context}
"""
对话要求：
1. 背景知识是最新的实时的信息，使用背景知识逐步思考、专业地回答问题。
2. 优先使用背景知识的内容回答我的问题，答案应与背景知识严格一致。
3. 背景知识无法回答我的问题时，可以忽略背景知识，根据你的知识来自由回答。
4. 使用对话的风格，自然的回答问题。
'''
strict_system = '''
你的背景知识:
"""
{context}
"""
对话要求：
1. 背景知识是最新的实时的信息，是你的唯一信息来源，使用背景知识回答问题。
2. 优先使用背景知识回答我的问题，答案与背景知识完全一致，无需做其他回答。
3. 背景知识与问题无关，或背景知识无法回答本次问题时，则拒绝回答本次问题：“亲，我不太清楚xxx”。
4. 使用对话的风格，自然的回答问题。
'''

# 我的问题是:"{question}"
template = ChatPromptTemplate.from_messages([
    ("system", system.strip()),
    # ("human", "Hello, how are you doing?"),
    # ("ai", "I'm doing well, thanks!"),
    ("human", "{question}"),
])

strict_template = ChatPromptTemplate.from_messages([
    ("system", strict_system.strip()),
    ("human", "{question}"),
])

if __name__ == '__main__':
    from meutils.pipe import *
    import langchain

    langchain.debug = True
    from langchain.chat_models import ChatOpenAI
    from langchain.chains import LLMChain

    llm = ChatOpenAI()

    context = """
2022年的某一天，李明开着摩托车去给客户送货，路上遇到了一只小狗，他停下车去看了一下小狗，回去开车的时候货物不见了。李明在2023年进了一批货物，货物里面居然有一只小狗。
    """
    c = LLMChain(llm=llm, prompt=template)
    # c = LLMChain(llm=llm, prompt=strict_template)
    # print(c.run(context=context, question="李明开摩托车遇到了什么动物？"))
    print(c.run(context=context, question="今天天气怎样"))
    print(type(strict_template))



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# @Project      : AI.  @by PyCharm
# @File         : client
# @Time         : 2023/5/30 16:39
# @Author       : betterme
# @WeChat       : meutils
# @Software     : PyCharm
# @Description  :

from meutils.pipe import *
import openai


def completion(
    prompt='你好', history=None, chat=False,
    api_base='https://api.openai-proxy.com/v1',  # https://t.me/openai_proxy
    api_key='sk-'
):
    openai.api_base = api_base
    openai.api_key = api_key
    kwargs = {
        "model": "gpt-3.5-turbo",
        "stream": True,
        "max_tokens": 1000,
        "temperature": None,
        "top_p": None,

        "user": "Betterme"
    }
    if is_open('0.0.0.0:8000'):
        logger.debug("本地大模型")
        openai.api_base = 'http://0.0.0.0:8000/v1'
        openai.api_key = 'chatllm'

    if chat:
        history = history or []  # [{"role": "system", "content": "你是东北证券大模型"}]
        kwargs['messages'] = history + [{"role": "user", "content": prompt}]
        completion = openai.ChatCompletion.create(**kwargs)
        response = ''
        for c in completion:
            _ = c.choices[0].get('delta').get('content', '')
            response += _
            print(_, flush=True, end='')
        # print('\n', response)

    else:
        kwargs['prompt'] = prompt
        kwargs['model'] = "text-davinci-003"
        completion = openai.Completion.create(**kwargs)
        response = ''
        for c in completion:
            _ = c.choices[0].text
            response += _
            print(_, flush=True, end='')
    return response


if __name__ == '__main__':
    # completion(prompt='你是谁', chat=True)
    completion(prompt='你是谁', chat=True)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# @Project      : AI.  @by PyCharm
# @File         : xunfei
# @Time         : 2023/7/24 13:46
# @Author       : betterme
# @WeChat       : meutils
# @Software     : PyCharm
# @Description  : https://www.xfyun.cn/doc/spark/Web.html


from pydantic import model_validator
from langchain.chat_models.openai import ChatOpenAI
from langchain.utils import get_from_dict_or_env, get_pydantic_field_names

from meutils.pipe import *
from chatllm.llmchain.completions import SparkBotCompletion


class SparkBot(ChatOpenAI):
    """
    api_key = {APP Id}:{API Key}:{Secret Key}
    """
    client: Any  #: :meta private:
    model_name: str = Field(default="spark-turbo", alias="model")
    openai_api_key: Optional[str] = Field(default=None, alias="spark_api_key")  # ernie_api_key: Optional[str] = None

    class Config:
        """Configuration for this pydantic object."""

        allow_population_by_field_name = True

    @model_validator(mode='after')
    def validate_environment(cls, values: Dict) -> Dict:
        """Validate that api key and python package exists in environment."""
        # 覆盖 openai_api_key
        values["openai_api_key"] = get_from_dict_or_env(
            values, "spark_api_key", "SPARK_API_KEY"
        )

        values["client"] = SparkBotCompletion

        if values["n"] < 1:
            raise ValueError("n must be at least 1.")
        if values["n"] > 1 and values["streaming"]:
            raise ValueError("n must be 1 when streaming.")
        return values

    @property
    def _llm_type(self) -> str:
        return Path(__file__).name  # 'ernie'


if __name__ == '__main__':
    from meutils.pipe import *
    from chatllm.llmchain.llms import SparkBot

    from langchain.chains import LLMChain
    from langchain.prompts import ChatPromptTemplate
    from langchain.callbacks import get_openai_callback

    first_prompt = ChatPromptTemplate.from_template("{q}")

    llm = SparkBot(streaming=True)
    # with get_openai_callback() as cb:
    #     c = LLMChain(llm=llm, prompt=first_prompt)
    #     print(c.run('你是谁'))
    #
    # print(cb.total_tokens)
    c = LLMChain(llm=llm, prompt=first_prompt)
    for i in c.run('你是谁'):
        print(i, end='')



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# @Project      : AI.  @by PyCharm
# @File         : copilot
# @Time         : 2023/12/6 13:14
# @Author       : betterme
# @WeChat       : meutils
# @Software     : PyCharm
# @Description  :

from meutils.pipe import *
from meutils.cache_utils import ttl_cache
from meutils.decorators.retry import retrying

from chatllm.llmchain.utils import tiktoken_encoder
from chatllm.schemas.kimi.protocol import EventData, State

from openai.types.chat import chat_completion_chunk, chat_completion
from openai.types.chat.chat_completion import ChatCompletion
from openai.types.chat.chat_completion_chunk import ChatCompletionChunk

requests.get = retrying(requests.get)
requests.post = retrying(requests.post)


class Completions(object):
    def __init__(self, **client_params):
        self.client_params = client_params
        api_key = self.client_params.get('api_key')
        self.access_token = self.get_access_token(api_key)

    def create(self, messages: List[Dict[str, Any]], **kwargs):
        """
                messages=messages,
                model=model,
                stream=stream
        :param messages:
        :param kwargs:
        :return:
        """

        data = {
            "model": "gpt-4",
            "messages": messages,
            "stream": self.client_params.get('stream'),
            **kwargs
        }

        if self.client_params.get('stream'):
            return self._stream_create(**data)
        else:
            return self._create(**data)

    def _create(self, **data):
        data = data or {
            "model": "gpt-4",
            "messages": [{"role": "user", "content": "你是谁"}]
        }
        response = self._post(**data)

        response = response.json()
        response['model'] = data.get('model', 'gpt-4')
        response['object'] = 'chat.completion'

        return ChatCompletion.model_validate(response)

    def _stream_create(self, **data):
        data = data or {
            "model": "gpt-4",
            "messages": [{"role": "user", "content": "你是谁"}],
            "stream": True
        }
        response = self._post(**data)

        buffer = b''  # 用于缓存不完整的行
        for chunk in response.iter_content(chunk_size=8096):  # , decode_unicode=True
            if chunk:
                buffer += chunk
                lines = buffer.split(b'\n\n')  # 按照两个换行符切分
                buffer = lines.pop()  # 最后一行可能不完整，保存到 buffer 中
                for line in lines:

                    # logger.debug(line)

                    if line == b'data: [DONE]':
                        break

                    line = line.strip(b"data: ")  # line.decode('utf-8').strip("data: ")
                    line = json.loads(line)
                    line['model'] = data.get('model', "gpt-4")
                    line['object'] = "chat.completion.chunk"
                    line['choices'][0]['finish_reason'] = line['choices'][0].get('finish_reason')  # 最后一个应为 "stop"

                    line = ChatCompletionChunk.model_validate(line)
                    line.choices[0].delta.role = 'assistant'

                    # logger.debug(line)

                    yield line

    def _post(self, **data):
        headers = {
            'Authorization': f'Bearer {self.access_token}',
            # 'X-Request-Id': str(uuid.uuid4()),
            # 'Vscode-Sessionid': str(uuid.uuid4()) + str(int(datetime.datetime.utcnow().timestamp() * 1000)),
            # 'vscode-machineid': machine_id,
            'Editor-Version': 'vscode/1.84.2',
            'Editor-Plugin-Version': 'copilot-chat/0.10.2',
            'Openai-Organization': 'github-copilot',
            'Openai-Intent': 'conversation-panel',
            'Content-Type': 'application/json',
            'User-Agent': 'GitHubCopilotChat/0.10.2',
            'Accept': '*/*',
            'Accept-Encoding': 'gzip, deflate, br',
        }

        response = requests.post(
            'https://api.githubcopilot.com/chat/completions',
            json=data,
            headers=headers,
            stream=data.get('stream')
        )

        return response

    @staticmethod
    @ttl_cache(ttl=60)
    def get_access_token(api_key: Optional[str] = None):
        api_key = api_key or os.getenv("GITHUB_COPILOT_TOKEN")
        assert api_key

        headers = {
            'Host': 'api.github.com',
            'authorization': f'token {api_key}',
            "Editor-Version": "vscode/1.84.2",
            "Editor-Plugin-Version": "copilot/1.138.0",
            "User-Agent": "GithubCopilot/1.138.0",
            "Accept": "*/*",
            "Accept-Encoding": "gzip, deflate, br",
            "Connection": "close"
        }
        response = requests.get('https://api.github.com/copilot_internal/v2/token', headers=headers)
        return response.json()['token']


if __name__ == '__main__':
    # print(Completions()._create())

    _ = Completions()._stream_create()

    for i in _:
        print(i)
